<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Zoom Online block
 *
 * @package    block_zoomonline
 * @copyright  2024 Ciaran Mac Donncha
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
defined('MOODLE_INTERNAL') || die();

/**
 * Handles the upgrade process for the block_zoomonline plugin.
 *
 * @param int $oldversion The version of the plugin before the upgrade.
 * @return bool True on successful upgrade.
 */
function xmldb_block_zoomonline_upgrade($oldversion) {
    global $DB;

    $dbman = $DB->get_manager(); // Database manager.

    if ($oldversion < 2025012601) {
        // Ensure the block_zoomonline_links table exists and is correct.
        $table = new xmldb_table('block_zoomonline_links');
        $fields = [
                new xmldb_field('id', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, XMLDB_SEQUENCE, null),
                new xmldb_field('zoom_meeting_id', XMLDB_TYPE_CHAR, '45', null, null, null, null),
                new xmldb_field('lecturer_id', XMLDB_TYPE_CHAR, '45', null, null, null, null),
                new xmldb_field('course_id', XMLDB_TYPE_CHAR, '45', null, null, null, null),
                new xmldb_field('crscode', XMLDB_TYPE_CHAR, '90', null, null, null, null),
                new xmldb_field('groupid', XMLDB_TYPE_CHAR, '90', null, null, null, null),
                new xmldb_field('visible', XMLDB_TYPE_INTEGER, '1', null, XMLDB_NOTNULL, null, '1'),
        ];

        if (!$dbman->table_exists($table)) {
            foreach ($fields as $field) {
                $table->add_field($field);
            }
            $table->add_key(new xmldb_key('id', XMLDB_KEY_PRIMARY, ['id']));
            $dbman->create_table($table);
        } else {
            foreach ($fields as $field) {
                if (!$dbman->field_exists($table, $field->getName())) {
                    $dbman->add_field($table, $field);
                }
            }
        }

        // Ensure the block_zoomonline_att_check table exists and is correct.
        $table = new xmldb_table('block_zoomonline_att_check');
        $fields = [
                new xmldb_field('id', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, XMLDB_SEQUENCE, null),
                new xmldb_field('zoom_meeting_id', XMLDB_TYPE_CHAR, '45', null, null, null, null),
                new xmldb_field('start_date', XMLDB_TYPE_DATETIME, null, null, null, null),
        ];

        if (!$dbman->table_exists($table)) {
            foreach ($fields as $field) {
                $table->add_field($field);
            }
            $table->add_key(new xmldb_key('id', XMLDB_KEY_PRIMARY, ['id']));
            $dbman->create_table($table);
        } else {
            foreach ($fields as $field) {
                if (!$dbman->field_exists($table, $field->getName())) {
                    $dbman->add_field($table, $field);
                }
            }
        }

        // Ensure the block_zoomonline_token table exists and is correct.
        $table = new xmldb_table('block_zoomonline_token');
        $fields = [
                new xmldb_field('id', XMLDB_TYPE_INTEGER, '10', null, XMLDB_NOTNULL, XMLDB_SEQUENCE, null),
                new xmldb_field('access_token', XMLDB_TYPE_CHAR, '1000', null, null, null, null),
                new xmldb_field('current', XMLDB_TYPE_INTEGER, '11', null, null, null, null),
        ];

        if (!$dbman->table_exists($table)) {
            foreach ($fields as $field) {
                $table->add_field($field);
            }
            $table->add_key(new xmldb_key('id', XMLDB_KEY_PRIMARY, ['id']));
            $dbman->create_table($table);
        } else {
            foreach ($fields as $field) {
                if (!$dbman->field_exists($table, $field->getName())) {
                    $dbman->add_field($table, $field);
                }
            }
        }

        // Upgrade savepoint - single call after all changes.
        upgrade_block_savepoint(true, 2025012601, 'zoomonline');
    }

    return true;
}
